/*
 * Licensed to Elastic Search and Shay Banon under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. Elastic Search licenses this
 * file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.elasticsearch.cluster.routing;

import org.elasticsearch.common.io.stream.StreamInput;
import org.elasticsearch.common.io.stream.StreamOutput;
import org.elasticsearch.common.io.stream.Streamable;
import org.elasticsearch.index.shard.ShardId;

import java.io.IOException;
import java.io.Serializable;

/**
 * Shard routing represents the state of a shard instance allocated in the cluster.
 *
 * @author kimchy (shay.banon)
 */
public interface ShardRouting extends Streamable, Serializable {

    /**
     * The shard id.
     */
    ShardId shardId();

    /**
     * The index name.
     */
    String index();

    /**
     * The index name.
     */
    String getIndex();

    /**
     * The shard id.
     */
    int id();

    /**
     * The shard id.
     */
    int getId();

    /**
     * The shard state.
     */
    ShardRoutingState state();

    /**
     * The shard is unassigned (not allocated to any node).
     */
    boolean unassigned();

    /**
     * The shard is initializing (usually recovering either from peer shard
     * or from gateway).
     */
    boolean initializing();

    /**
     * The shard is in started mode.
     */
    boolean started();

    /**
     * The shard is in relocating mode.
     */
    boolean relocating();

    /**
     * Relocating or started.
     */
    boolean active();

    /**
     * The shard is assigned to a node.
     */
    boolean assignedToNode();

    /**
     * The current node id the shard is allocated to.
     */
    String currentNodeId();

    /**
     * The relocating node id the shard is either relocating to or relocating from.
     */
    String relocatingNodeId();

    /**
     * Is this a primary shard.
     */
    boolean primary();

    /**
     * A short description of the shard.
     */
    String shortSummary();

    /**
     * A shard iterator with just this shard in it.
     */
    ShardsIterator shardsIt();

    /**
     * Does not write index name and shard id
     */
    void writeToThin(StreamOutput out) throws IOException;

    void readFromThin(StreamInput in) throws ClassNotFoundException, IOException;
}
